import click
import os
import json
from pathlib import Path
from .cmd_app import app_cmd_group

HOME = Path('~/.rxd').expanduser().resolve()
CONFIG_PATH = HOME.joinpath("config.json")


@click.group(name="rxd")
def main():
    pass


@main.command()
@click.option("-w", "--workspace",
              required=False,
              type=str,
              default=Path('~/workspace').expanduser().resolve(),
              help="Path to directory where your apps will be stored")
def setup(workspace):

    # create home directory
    if not HOME.exists():
        print(f"Making directory: {HOME}")
        os.makedirs(HOME)

    # write config file to point to workspace
    if not CONFIG_PATH.exists():
        with open(CONFIG_PATH, "w") as fh:
            print(f"Saving config: {CONFIG_PATH}")
            json.dump({"workspace": workspace}, fh, indent=4)

    # create workspace
    workspace_path = Path(workspace)
    if not workspace_path.exists():
        print(f"Making directory: {workspace_path}")
        workspace_path.mkdir()


@main.command()
def info():
    print(f"Home: {HOME}")
    print(f"Config file: {CONFIG_PATH}")
    if CONFIG_PATH.exists():
        with open(CONFIG_PATH) as fh:
            config = json.load(fh)
        workspace = Path(config['workspace'])\
            .expanduser()\
            .resolve()
        print(f"Workspace: {workspace}")


main.add_command(app_cmd_group)

if __name__ == "__main__":
    main()



import json
import sys
import getpass
import textwrap
import subprocess as sp
from pathlib import Path


def run(cmdline: list, ignore_errors=False):
    try:
        print("Running %s" % " ".join(cmdline))
        sp.check_call(cmdline)
    except sp.CalledProcessError:
        if not ignore_errors:
            raise


class Systemd:
    def __init__(self):
        pass

    @classmethod
    def is_available(cls):
        return Path('/etc/systemd/system').exists()

    @classmethod
    def install(cls, app):
        pass

    @classmethod
    def remove(cls, app):
        service_name = app.systemd_service_name
        install_path = Path(f'/etc/systemd/system/{service_name}')
        cls.stop(app)
        cls.disable(app)
        if install_path.exists():
            print("Removing %s" % install_path)
            run(["sudo", "rm", str(install_path)])

    @classmethod
    def start(cls, app):
        run(["sudo", "systemctl", "start",
             app.systemd_service_name],
            ignore_errors=True)

    @classmethod
    def stop(cls, app):
        run(["sudo", "systemctl", "stop",
             app.systemd_service_name],
            ignore_errors=True)

    @classmethod
    def restart(cls, app):
        run(["sudo", "systemctl", "restart",
            app.systemd_service_name], ignore_errors=True)
        pass

    @classmethod
    def enable(cls, app):
        run(["sudo", "systemctl", "enable",
            app.systemd_service_name], ignore_errors=True)

    @classmethod
    def disable(cls, app):
        run(["sudo", "systemctl", "disable",
            app.systemd_service_name], ignore_errors=True)

    @classmethod
    def logs(cls, app, follow):
        if follow:
            run(["journalctl", "-f", "-u", app.systemd_service_name],
                ignore_errors=True)
        else:
            run(["journalctl", "-e", "-u", app.systemd_service_name],
                ignore_errors=True)

    @classmethod
    def daemonize(cls, app):
        python_path = sys.executable
        user = getpass.getuser()
        systemd_service_file = \
            f"""
            [Unit]
            Description=rxd
            After=network.target
            StartLimitIntervalSec=30
            StartLimitBurst=5

            [Service]
            Type=simple
            User={user}
            ExecStart={python_path} -m rxd.cmd app run {app.name}
            Restart=always
            RestartSec=5

            [Install]
            WantedBy=default.target
            """
        systemd_service_file = textwrap.dedent(systemd_service_file)
        systemd_service_install_path = Path(
            f'/etc/systemd/system/{app.systemd_service_name}')
        print(f'Writing {app.systemd_services_definition_path}')
        with open(app.systemd_services_definition_path, "w") as fh:
            fh.write(systemd_service_file)

        if Path('/etc/systemd/system').exists():
            if not systemd_service_install_path.exists():
                print("Moving %s -> %s" %
                      (app.systemd_services_definition_path,
                       systemd_service_install_path))
                print("We need sudo permissions to do so")
                sp.check_call(
                    ["sudo", "mv",
                     app.systemd_services_definition_path,
                     systemd_service_install_path])
                sp.check_call(["sudo", "systemctl",
                               "daemon-reload"])
        else:
            print("Systemd does not exist")

    @classmethod
    def status(cls, app, output):
        if not Path('/etc/systemd/system').exists():
            return None

        lines = sp.check_output(["systemctl",
                                 "show",
                                 "--no-pager",
                                 app.systemd_service_name])
        lines = lines.decode("utf-8").split("\n")
        status_data = {}
        for line in lines:
            kv = line.strip().split("=", 1)
            if len(kv) == 2:
                key, value = kv
                status_data[key] = value

        status = dict(
            is_running=status_data.get('ActiveState') == 'active',
            pid=status_data.get('MainPID'),
            user=status_data.get('User'),
            memory_mb=status_data.get('MemoryCurrent'),
            state=status_data.get('UnitFileState')
        )
        if output:
            print(json.dumps(status, indent=4))
        return status



import os
import subprocess as sp
import json
import typing as t
from pathlib import Path
from .daemonizers import Systemd


HOME = Path('~/.rxd').expanduser().resolve()
CONFIG_PATH = HOME.joinpath("config.json")
APP_DIR = Path(HOME, "apps")


# Set workspace directory
WORKSPACE_DIR = Path('~/workspace').expanduser().resolve()
if CONFIG_PATH.exists():
    with open(CONFIG_PATH) as fh:
        config = json.load(fh)
        WORKSPACE_DIR = Path(config["workspace"]).expanduser().resolve()


class Chdir:

    def __init__(self, directory):
        self.directory = directory
        self.oldcwd = None

    def __enter__(self):
        print("changing directory to %s" % self.directory)
        self.oldcwd = Path(os.curdir).resolve()
        os.chdir(self.directory)

    def __exit__(self, type, value, traceback):

        if self.oldcwd is not None:
            print("changing directory back to %s" % self.oldcwd)
            os.chdir(self.oldcwd)
            self.oldcwd = None


class Manager:
    def __init__(self):
        pass

    def list(self):
        return [Application.load(app_file.split(".json")[0])
                for app_file in os.listdir(APP_DIR)]

    def add(self, name, repo):
        app = Application(name, repo)
        app.save()

    def remove(self, name):
        return Application.load(name).remove()

    def fetch(self, name):
        app = Application.load(name)
        app.fetch()


class Application:
    def __init__(self, name, repo=None, deploy_name=None) -> None:
        self.name = name
        self.repo = repo
        self.deploy_name = ".deploy" if not deploy_name else deploy_name

    @property
    def repo_name(self) -> t.Union[str, None]:
        if self.repo:
            return str(Path(self.repo).stem)
        return None

    @property
    def repo_path(self) -> t.Union[Path, None]:
        if self.repo and self.repo_name:
            return self.workspace_container_path\
                       .joinpath(self.repo_name)\
                       .resolve()
        return None

    @property
    def metadata_path(self) -> Path:
        return Path(HOME, "apps", "%s.json" % self.name).resolve()

    @property
    def workspace_container_path(self) -> Path:
        return Path(WORKSPACE_DIR, self.name).resolve()

    @property
    def systemd_services_definition_path(self) -> Path:
        return self.workspace_container_path.joinpath("app.service").resolve()

    @property
    def systemd_service_name(self):
        return f"rxd-app-{self.name}.service"

    def save(self):
        self.setup_metadata()
        path = self.metadata_path
        with open(path, "w") as fh:
            json.dump({
                "name": self.name,
                "repo": self.repo,
                "deploy_name": self.deploy_name
            }, fh)

    def setup_metadata(self):
        if not self.metadata_path.parent.exists():
            os.makedirs(self.metadata_path.parent)

    def setup_workspace(self):
        if not self.workspace_container_path.exists():
            os.makedirs(self.workspace_container_path)

    def exists(self):
        return self.metadata_path.exists()

    @classmethod
    def load(cls, name):
        metadata_path = Path(HOME,
                             "apps",
                             "%s.json" % name).resolve()

        if metadata_path.exists():
            with open(metadata_path, "r") as fh:
                data = json.load(fh)
                return Application(name=data['name'],
                                   repo=data['repo'],
                                   deploy_name=data.get('deploy_name'))
        else:
            return Application(name=name)

    def fetch(self):
        self.setup_workspace()

        # if no repo
        if self.workspace_container_path.exists() \
                and self.repo_path \
                and not self.repo_path.exists():
            with Chdir(self.workspace_container_path):
                sp.check_call("git clone --depth 1 %s" % self.repo,
                              shell=True)

        # if repo exists
        elif self.workspace_container_path.exists()\
                and self.repo_path \
                and self.repo_path.exists():
            with Chdir(self.repo_path):
                sp.check_call("git pull origin main",
                              shell=True)

    def remove(self):
        if workspace_path := self.workspace_container_path:
            answer = input("Are you sure you want to remove '%s' [y/n] ?: "
                           % workspace_path)
            if answer.strip().lower() == 'y':
                Systemd.remove(self)

                # remove app's workspace folder
                if workspace_path.exists():
                    print("Removing %s" % workspace_path)
                    import shutil
                    shutil.rmtree(workspace_path)

                # remove app's metadata file
                if metadata_path := self.metadata_path:
                    if self.metadata_path.exists():
                        print("Removing %s" % metadata_path)
                        metadata_path.unlink()

    def build(self):
        if repo_path := self.repo_path:
            if not repo_path.exists():
                print(f"{repo_path} does not exist, please run "
                      f"'rxd app fetch {self.name}'")
                return
            with Chdir(repo_path):
                sp.check_call(["/usr/bin/bash", f"{self.deploy_name}/build"],
                              shell=True)

    def run(self):
        if repo_path := self.repo_path:
            if not repo_path.exists():
                print(f"{repo_path} does not exist, please run "
                      f"'rxd app fetch {self.name}'")
                return
            with Chdir(repo_path):
                sp.check_call(["/usr/bin/bash", f"{self.deploy_name}/run"],
                              shell=True)

    def daemonize(self):
        Systemd.daemonize(self)

    def status(self, output=False):
        return Systemd.status(self, output=output)

    def start(self):
        Systemd.start(self)

    def stop(self):
        Systemd.stop(self)

    def restart(self):
        Systemd.restart(self)

    def enable(self):
        Systemd.enable(self)

    def disable(self):
        Systemd.disable(self)

    def logs(self, follow=False):
        Systemd.logs(self, follow=follow)

    def __repr__(self):
        return "Application(name=%s, repo=%s)" % (self.name, self.repo)
